#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <time.h>
#include <assert.h>
#include "hashtable.h"
#include "hashtable_itr.h"

#define _DEBUG

#define min(a,b)        (((a)<(b))?(a):(b))
#define max(a,b)        (((a)>(b))?(a):(b))

#define false 0
#define true 1

#define MAXLEN 20000

struct key {
    char keyword[64]; 
};

struct value {
    float weight;
};

DEFINE_HASHTABLE_INSERT(insert_some, struct key, struct value);
DEFINE_HASHTABLE_SEARCH(search_some, struct key, struct value);
DEFINE_HASHTABLE_REMOVE(remove_some, struct key, struct value);
DEFINE_HASHTABLE_ITERATOR_SEARCH(search_itr_some, struct key);

struct hashtable *h[3], *x;

/*****************************************************************************/
static unsigned int
hashfromkey(void *ky)
{
    struct key *k = (struct key *)ky;
    char *str = k->keyword;

    unsigned int hash = 0;
    int c;

    while ((c = *str++))
        hash = c + (hash << 6) + (hash << 16) - hash;

    return hash;
}

static int
equalkeys(void *k1, void *k2)
{
    struct key *kk1 = k1;
    struct key *kk2 = k2;
    return (0 == strcmp(kk1->keyword,kk2->keyword));
    //  return (0 == memcmp(k1,k2,sizeof(struct key)));
}

/*********************************************************/
float coin() {
    return rand() * pow(RAND_MAX + 1.0, -1);
}

int krand(int k) {
    return rand() % k;
}

int sisspace(char *s) {
    while((*s)!='\0') {
        if (!isspace(*s)) return 0;
        s++;
    }
    return 1;
}

int dot_product(int inner_node, struct hashtable *x) {

    float ret = 0;
    struct hashtable_itr *itr_x = hashtable_iterator(x);
    struct key *k;
    struct value *v, *found;

    if (hashtable_count(x) == 0) return;
    do {
        k = hashtable_iterator_key(itr_x);
        v = hashtable_iterator_value(itr_x);

        if (found = search_some(h[inner_node],k)) { /* k is in */
            ret += v->weight * found->weight;
        }
    } while(hashtable_iterator_advance(itr_x));

    return (ret >=0) ? 1 : -1;
}

/**********************************************************/

void update (int inner_node, struct hashtable *x, int binary_label) {
    struct hashtable_itr *itr_x = hashtable_iterator(x);
    struct key *k, *kk;
    struct value *v, *vv, *found;

    if ( dot_product(inner_node,x) == binary_label ) return;

    if (hashtable_count(x) == 0) return;
    do {

	k = hashtable_iterator_key(itr_x);
	v = hashtable_iterator_value(itr_x);

	if (NULL == (found = search_some(h[inner_node],k))) { /* k not in */
            vv = (struct value *)malloc(sizeof(struct value));
	    vv->weight = binary_label * v->weight;

	    kk = (struct key *)malloc(sizeof(struct key));
            if (NULL == kk) {
                printf("ran out of memory allocating a key\n");
                exit(-1);
            }
            strcpy(kk->keyword,k->keyword);

            if (!insert_some(h[inner_node],kk,vv)) exit(-1);
#ifdef DEBUG
	    printf("Inserted (%s,%2.3f)\n",kk->keyword,vv->weight);
#endif
        }
	else {  /* update the count of k in h[i] */
	    found->weight += binary_label * v->weight;
#ifdef DEBUG
	    printf("Added to %s (total count %2.3f)\n",
		k->keyword,found->weight);
#endif
	}
    } while(hashtable_iterator_advance(itr_x));

}

void print_hashtable (struct hashtable *w) {
    struct hashtable_itr *itr = hashtable_iterator(w);
    struct key *k;
    struct value *v;

    if (hashtable_count(w) == 0) {
#ifdef DEBUG
	printf("Hashtable is empty\n");  fflush(stdout);
#endif
	return;
    }

    do {

        k = hashtable_iterator_key(itr);
	v = hashtable_iterator_value(itr);
	printf("%s:%2.4f\n",k->keyword,v->weight);
	fflush(stdout);

    } while (hashtable_iterator_advance(itr));
}

int main (int argc, char *argv[]) {
    FILE *fp;
    int i, j, y, action, inner_node, freq, m;
    float Z;
    char *line, *example, *word;
    
    if (argc < 3) {
        fprintf(stderr,"Usage: %s data schedule\n",argv[0]);
        exit(1);
    }

    float mult = atof(argv[2]);

    if ((fp = fopen(argv[1],"r")) == NULL) { exit(1); }

    time_t tm;
    (void) time(&tm);
    srand(tm);
    
    int n = 0; 
    int err = 0; 
    int tree_prediction;

    int l, len;
    line = (char *)malloc(MAXLEN*sizeof(char));
    example = (char *)malloc(MAXLEN*sizeof(char));

    struct key *k, *kk;
    struct value *v, *found;

    float dot[3], gamma = 1;

    for(i=0; i<3; i++) {
        h[i] = create_hashtable(16, hashfromkey, equalkeys);
        if (NULL == h[i]) exit(-1); 
    }

    k = (struct key *)malloc(sizeof(struct key));
    if (NULL == k) {
        printf("ran out of memory allocating a key\n");
        return 1;
    }

    x = NULL;

    while ((l = getline(&line, &len, fp)) != -1 ){
        if (sisspace(line) || line[0]=='%' || line[0]=='@') { continue; }
	strcpy(example,line);

	y = atoi(strtok(line," :\t\n"));  
	Z = 0;  for(i=0; i<3; i++) dot[i] = 0;

	word = strtok(NULL," :\t\n");
	assert(0 == strcmp(word,"|features"));

#ifdef DEBUG
	printf("\nRead line %d (label %d)\n",n,y);  fflush(stdout);
#endif
	
	m = 0;
	/* receive x and make offset tree predictions */
	while (word = strtok(NULL," :\t\n")) {
	    freq = atoi(strtok(NULL," :\t\n"));
	    Z += freq*freq; m ++;

	    /* 0 - root, 1 - left child, 2 - right child */
	    strcpy(k->keyword,word);

	    for(i=0; i<3; i++) { 
  	        if (NULL != (found = search_some(h[i],k))) {
		    dot[i] += found->weight*freq;
                }
	    }
	}

	// get the prediction
	if ( dot[0] >= 0 ) {
	    if (dot[2] >= 0) tree_prediction = 3;
	    else tree_prediction = 2;
	}
	else {
	    if (dot[1] >= 0) tree_prediction = 1;
            else tree_prediction = 0;
	}

	// randomize the choice with bias gamma
	int explore = false; 
	if (coin() < gamma) { explore = true; action = krand(4); }
	else action = tree_prediction;

#ifdef DEBUG
        printf("Line %d, label %d, predicted label %d\n",
		n, y, action); fflush(stdout);
        printf("After insertion, x contains %u words (%d)\n",
            hashtable_count(x), m);
#endif

	if (y != action) {
	    err++;
	}

	if (1) {
            /* create x */
            if (x == NULL) {
                x = create_hashtable(16, hashfromkey, equalkeys);
                if (NULL == x) exit(-1);
#ifdef DEBUG
                printf("Creating x with %u items\n", hashtable_count(x));
                fflush(stdout);
#endif
            }

	    strtok(example," :\t\n");
            word = strtok(NULL," :\t\n");
            assert(0 == strcmp(word,"|features"));

            /* put the example into x */
            while (word = strtok(NULL," :\t\n")) {
                kk = (struct key *)malloc(sizeof(struct key));
                    if (NULL == kk) {
                    printf("ran out of memory allocating a key\n");
                    return 1;
                }
                freq = atoi(strtok(NULL," :\t\n"));
                strcpy(kk->keyword, word);
                v = (struct value *)malloc(sizeof(struct value));
                v->weight = (float) freq/Z;

#ifdef DEBUG2
                printf("Inserting %s (%f) into x\n",kk->keyword,v->weight);
#endif
                if (!insert_some(x,kk,v)) {
                    printf("out of memory inserting into x\n");
                    return 1;
                }
            }

	    inner_node = (action > 1) ? 2 : 1;

	    int binary_label;

	    if (y == action) // reward is 1
		binary_label = (action % 2 == 1) ? 1 : -1;
	    else { // reward is 0
		binary_label = (action % 2 == 1) ? -1 : 1;
	    }

	    update(inner_node, x, binary_label);

	    // update the dot product of the innner node
	    int binary_dot = dot_product(inner_node, x);

	    if (binary_dot == ((action % 2) ? 1 : -1 )) {
#ifdef DEBUG
	        printf("Root hashtable before update: %d items\n",
                    hashtable_count(h[0]));  fflush(stdout);
	        print_hashtable(h[0]);
#endif
		if (y == action) {
		    binary_label = (action / 2 == 1) ? 1 : -1;
		}
		else {
		    binary_label = (action / 2 == 1) ? -1 : 1;
		}
	        update(0, x, binary_label);
#ifdef DEBUG
                printf("Root hashtable after update: %d items\n",
		    hashtable_count(h[0]));
	        print_hashtable(h[0]);
#endif
	    }

	    hashtable_destroy(x,1);
	    x = NULL;
	}
	n++;
        gamma = mult*pow(n,-0.66);
	if (!(n%30)) printf("%d %d\n",n,err);
    }
    printf("%d %d\n",n,err);
}
