#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <time.h>
#include <assert.h>
#include "hashtable.h"
#include "hashtable_itr.h"

#define DEBUG_
#define DEBUG2_

#define min(a,b)        (((a)<(b))?(a):(b))
#define max(a,b)        (((a)>(b))?(a):(b))

#define MAXLEN 20000

struct key {
    char keyword[64]; 
};

struct value {
    float count;
};

DEFINE_HASHTABLE_INSERT(insert_some, struct key, struct value);
DEFINE_HASHTABLE_SEARCH(search_some, struct key, struct value);
DEFINE_HASHTABLE_REMOVE(remove_some, struct key, struct value);
DEFINE_HASHTABLE_ITERATOR_SEARCH(search_itr_some, struct key);

struct hashtable **h, *x;

/*****************************************************************************/
static unsigned int
hashfromkey(void *ky)
{
    struct key *k = (struct key *)ky;
    char *str = k->keyword;

    unsigned int hash = 0;
    int c;

    while ((c = *str++))
        hash = c + (hash << 6) + (hash << 16) - hash;

    return hash;
}

static int
equalkeys(void *k1, void *k2)
{
    // return (0 == memcmp(k1,k2,sizeof(struct key)));
    struct key *kk1 = k1;
    struct key *kk2 = k2;
    // printf("COMPARING '%s' '%s'\n",kk1->keyword,kk2->keyword);

    return (0 == strcmp(kk1->keyword,kk2->keyword));
}

/*****************************************************************************/

float coin() {
    return rand() * pow(RAND_MAX + 1.0, -1);
}

int krand(int k) {
    return rand() % k;
}

int sisspace(char *s) {
    while((*s)!='\0') {
        if (!isspace(*s)) return 0;
        s++;
    }
    return 1;
}

void update (int i, struct hashtable *x, int y, int Z) {

    struct hashtable_itr *itr_x = hashtable_iterator(x);
    struct key *k, *kk;
    struct value *v, *vv, *found;

    if (hashtable_count(x) == 0) return;
    do {

	k = hashtable_iterator_key(itr_x);
	v = hashtable_iterator_value(itr_x);

	if (NULL == (found = search_some(h[i],k))) { /* k not in h[i] */
            vv = (struct value *)malloc(sizeof(struct value));
	    vv->count = y * v->count*pow(Z,-1);

	    kk = (struct key *)malloc(sizeof(struct key));
            if (NULL == kk) {
                printf("ran out of memory allocating a key\n");
                exit(-1);
            }
	    strcpy(kk->keyword,k->keyword);

            if (!insert_some(h[i],kk,vv)) exit(-1);
#ifdef DEBUG2
	    printf("Inserted (%s,%2.1f)\n",kk->keyword,y*v->count);
	    fflush(stdout);
#endif
        }
	else {  /* update the count of k in h[i] */
	    found->count += y * v->count*pow(Z,-1);
#ifdef DEBUG2
	    printf("Added to %s (%2.1f)\n",k->keyword,y*v->count);
	    fflush(stdout);
#endif
	}
    } while(hashtable_iterator_advance(itr_x));

    if (y == -1) {
#ifdef DEBUG2
        printf("Before destroying\n"); fflush(stdout);
#endif
        hashtable_destroy(x,1);
#ifdef DEBUG2
        printf("After destroying\n"); fflush(stdout);
#endif
    }
}

void print_hashtable (struct hashtable *w) {
    struct hashtable_itr *itr = hashtable_iterator(w);
    struct key *k;
    struct value *v;

    if (hashtable_count(w) == 0) {
#ifdef DEBUG
	printf("Hashtable is empty\n");  fflush(stdout);
#endif
	return;
    }

    do {

        k = hashtable_iterator_key(itr);
	v = hashtable_iterator_value(itr);
	printf("%s:%2.4f\n",k->keyword,v->count);
	fflush(stdout);

    } while (hashtable_iterator_advance(itr));
}

int main (int argc, char *argv[]) {
    FILE *fp;

    int K = 4;
    int i, j, y, a, freq, Z;

    char *line, *example, *word;
    
    if (argc < 2) {
        fprintf(stderr,"Usage: %s data\n",argv[0]);
        exit(1);
    }

    if ((fp = fopen(argv[1],"r")) == NULL) { exit(1); }

    time_t tm;
    (void) time(&tm);
    srand(tm);
    
    int n = 0; 
    int err = 0; 

    int l, len, predict;
    line = (char *)malloc(MAXLEN*sizeof(char));
    example = (char *)malloc(MAXLEN*sizeof(char));

    struct key *k, *kk;
    struct value *v, *found;
    struct hashtable_itr *itr;

    float *dot = (float *)malloc(sizeof(float)*K);

    h = (struct hashtable **)malloc(sizeof(struct hashtable *)*K);

    for(i=0; i<K; i++) {
        h[i] = create_hashtable(16, hashfromkey, equalkeys);
        if (NULL == h[i]) exit(-1); 
    }

    k = (struct key *)malloc(sizeof(struct key));
    if (NULL == k) {
	printf("ran out of memory allocating a key\n");
	return 1;
    }

    x = NULL;

    while ((l = getline(&line, &len, fp)) != -1 ){
        if (sisspace(line) || line[0]=='%' || line[0]=='@') { continue; }
	strcpy(example,line);

	y = atoi(strtok(line," :\t\n"));  
	Z = 0;  for(i=0; i<K; i++) dot[i] = 0;

	word = strtok(NULL," :\t\n");
	assert(0 == strcmp(word,"|features"));

#ifdef DEBUG
	printf("\nLine %d (label %d)\n",n,y);  fflush(stdout);
#endif
	
	/* receive x and make offset tree predictions */
	while (word = strtok(NULL," :\t\n")) {
	    freq = atoi(strtok(NULL," :\t\n"));
	    Z += freq*freq; 

	    /* search for word in h, update the dot products */
	    strcpy(k->keyword,word);
#ifdef DEBUG
	    printf("searching for word %s\n",k->keyword);
#endif
	    for(i=0; i<K; i++) { 
		// print_hashtable(h[i]);
  	        if (NULL != (found = search_some(h[i],k))) {
		    dot[i] += found->count*freq;
#ifdef DEBUG
		    printf("found in %d\n",i);
#endif
                }
	    }
	}

#ifdef DEBUG
        printf("Dot products: ");
	for(i=0; i<K; i++) {
            printf("%3.2f ", dot[i]);
        }
	printf("\n");
#endif

	/* compute the prediction */
	predict = 0; a = 0;
	for(i=0; i<K; i++) {
	    if (dot[i] > predict) {
		predict = dot[i]; a = i;
	    }
	}

#ifdef DEBUG
	if (y != a) printf("Mistake--------------------\n");
        printf("label %d\t predicted label %d\n",y,a); 
	fflush(stdout);
#endif

	// update the hyperplanes on (x,a,I(y=a))
	// if y = a, there is no update
	if (y != a) {

	    /* create x */
            if (x == NULL) {
                x = create_hashtable(16, hashfromkey, equalkeys);
                if (NULL == x) exit(-1);
#ifdef DEBUG
                printf("Creating x with %u items\n", hashtable_count(x));
		fflush(stdout);
#endif
            }

	    strtok(example," :\t\n");
            word = strtok(NULL," :\t\n");
            assert(0 == strcmp(word,"|features"));

	    /* put the example into x */
	    while (word = strtok(NULL," :\t\n")) {
		kk = (struct key *)malloc(sizeof(struct key));
	            if (NULL == kk) {
       	            printf("ran out of memory allocating a key\n");
                    return 1;
                }
                freq = atoi(strtok(NULL," :\t\n"));
                strcpy(kk->keyword, word);
                v = (struct value *)malloc(sizeof(struct value));
                v->count = (float) freq;

#ifdef DEBUG2
		printf("Inserting %s (%f) into x\n",kk->keyword,v->count);
#endif
                if (!insert_some(x,kk,v)) {
                    printf("out of memory inserting into x\n");
                    return 1;
                }
	    }

#ifdef DEBUG
	    printf("x contains %d items\n",hashtable_count(x));
	    for(i=0; i<K; i++) {
	        printf("h[%d] contains %d items\n",i,hashtable_count(h[i]));
	    }
#endif

#ifdef DEBUG
	    printf("Add x to h[%d]\n",y);
#endif 
	    update(y, x, 1, Z);
#ifdef DEBUG
	    for(i=0; i<K; i++) {
                printf("h[%d] contains %d items\n",i,hashtable_count(h[i]));
            }
#endif


#ifdef DEBUG
	    printf("Subtract x from h[%d]\n",a);
#endif
	    update(a, x, -1, Z);
#ifdef DEBUG
            for(i=0; i<K; i++) {
                printf("h[%d] contains %d items\n",i,hashtable_count(h[i]));
            }
#endif
	    x = NULL;
	    err++;
	}
	n++;
	if (!(n%30)) printf("%d %d\n",n,err); fflush(stdout);
	// printf("%d %d\n",n,err);
    }
    printf("%d %d\n",n,err);
}
